from __future__ import annotations

import logging
import os
import random
import tempfile
import warnings
import xml.etree.ElementTree as et
import zipfile
from itertools import chain

import numpy

# type: ignore
# pylint: disable=unused-variable, too-many-arguments, consider-using-with, no-else-return, too-many-nested-blocks, redundant-u-string-prefix

# Copyright (C) 2014 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html


r"""Python wrapper for `Latent Dirichlet Allocation (LDA) <https://en.wikipedia.org/wiki/Latent_Dirichlet_allocation>`_
from `MALLET, the Java topic modelling toolkit <http://mallet.cs.umass.edu/>`_

This module allows both LDA model estimation from a training corpus and inference of topic distribution on new,
unseen documents, using an (optimized version of) collapsed gibbs sampling from MALLET.

Notes
-----
MALLET's LDA training requires :math:`O(corpus\_words)` of memory, keeping the entire corpus in RAM.
If you find yourself running out of memory, either decrease the `workers` constructor parameter,
or use :class:`gensim.models.ldamodel.LdaModel` or :class:`gensim.models.ldamulticore.LdaMulticore`
which needs only :math:`O(1)` memory.
The wrapped model can NOT be updated with new documents for online training -- use
:class:`~gensim.models.ldamodel.LdaModel` or :class:`~gensim.models.ldamulticore.LdaMulticore` for that.

Installation
------------
Use `official guide <http://mallet.cs.umass.edu/download.php>`_ or this one ::

    sudo apt-get install default-jdk
    sudo apt-get install ant
    git clone git@github.com:mimno/Mallet.git
    cd Mallet/
    ant

Examples
--------
.. sourcecode:: pycon

    >>> from gensim.test.utils import common_corpus, common_dictionary
    >>> from gensim.models.wrappers import LdaMallet
    >>>
    >>> path_to_mallet_binary = "/path/to/mallet/binary"
    >>> model = LdaMallet(path_to_mallet_binary, corpus=common_corpus, num_topics=20, id2word=common_dictionary)
    >>> vector = model[common_corpus[0]]  # LDA topics of a documents

"""

try:
    import gensim.matutils as matutils
    import gensim.models.basemodel as basemodel
    import gensim.utils as utils
    from gensim.models.ldamodel import LdaModel
    from gensim.utils import check_output, revdict
except ImportError:
    ...

logger = logging.getLogger(__name__)


class LdaMallet(utils.SaveLoad, basemodel.BaseTopicModel):
    """Python wrapper for LDA using `MALLET <http://mallet.cs.umass.edu/>`_.

    Communication between MALLET and Python takes place by passing around data files on disk
    and calling Java with subprocess.call().

    Warnings
    --------
    This is **only** python wrapper for `MALLET LDA <http://mallet.cs.umass.edu/>`_,
    you need to install original implementation first and pass the path to binary to ``mallet_path``.

    """

    def __init__(
        self,
        mallet_path: str,
        corpus=None,
        num_topics: int = 100,
        alpha: float = 50,
        id2word: dict = None,
        workers: int = 4,
        prefix: str = None,
        optimize_interval: int = 0,
        iterations: int = 1000,
        topic_threshold: float = 0.005,
        random_seed: int = 0,
        num_top_words: int = 1000,
    ):
        """

        Parameters
        ----------
        mallet_path : str
            Path to the mallet binary, e.g. `/home/username/mallet-2.0.7/bin/mallet`.
        corpus : iterable of iterable of (int, int), optional
            Collection of texts in BoW format.
        num_topics : int, optional
            Number of topics.
        alpha : int, optional
            Alpha parameter of LDA.
        id2word : :class:`~gensim.corpora.dictionary.Dictionary`, optional
            Mapping between tokens ids and words from corpus, if not specified - will be inferred from `corpus`.
        workers : int, optional
            Number of threads that will be used for training.
        prefix : str, optional
            Prefix for produced temporary files.
        optimize_interval : int, optional
            Optimize hyperparameters every `optimize_interval` iterations
            (sometimes leads to Java exception 0 to switch off hyperparameter optimization).
        iterations : int, optional
            Number of training iterations.
        topic_threshold : float, optional
            Threshold of the probability above which we consider a topic.
        random_seed: int, optional
            Random seed to ensure consistent results, if 0 - use system clock.

        """
        self.mallet_path: str = mallet_path
        self.id2word: dict[int, str] = id2word
        if self.id2word is None:
            logger.warning("no word id mapping provided; initializing from corpus, assuming identity")
            self.id2word = utils.dict_from_corpus(corpus)
            self.num_terms = len(self.id2word)
        else:
            self.num_terms: int = 0 if not self.id2word else 1 + max(self.id2word.keys())
        if self.num_terms == 0:
            raise ValueError("cannot compute LDA over an empty collection (no terms)")
        self.num_topics: int = num_topics
        self.topic_threshold: float = topic_threshold
        self.alpha: float = alpha
        if prefix is None:
            rand_prefix = hex(random.randint(0, 0xFFFFFF))[2:] + '_'
            prefix = os.path.join(tempfile.gettempdir(), rand_prefix)
        self.prefix: str = prefix
        self.workers: int = workers
        self.optimize_interval: int = optimize_interval
        self.iterations: int = iterations
        self.num_top_words: int = num_top_words
        self.random_seed: int = random_seed
        if corpus is not None:
            self.train(corpus)

    def inferencer_filename(self):
        """Get path to inferencer.mallet file."""
        return self.prefix + 'inferencer.mallet'

    def topic_keys_filename(self):
        """Get path to topic keys text file."""
        return self.prefix + 'topickeys.txt'

    def mallet_state_filename(self):
        """Get path to temporary file."""
        return self.prefix + 'state.mallet.gz'

    def document_topics_filename(self):
        """Get path to document topic text file."""
        return self.prefix + 'doctopics.txt'

    def text_corpus_filename(self):
        """Get path to corpus text file."""
        return self.prefix + 'corpus.txt'

    def mallet_corpus_filename(self):
        """Get path to corpus.mallet file."""
        return self.prefix + 'corpus.mallet'

    def word_weights_filename(self):
        """Get path to word weight file."""
        return self.prefix + 'wordweights.txt'

    def corpus2mallet(self, corpus, file_like):
        """Convert `corpus` to Mallet format and write it to `file_like` descriptor.

        Format ::

            document id[SPACE]label (not used)[SPACE]whitespace delimited utf8-encoded tokens[NEWLINE]

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Collection of texts in BoW format.
        file_like : file-like object
            Opened file.

        """
        for docno, doc in enumerate(corpus):
            if self.id2word:
                tokens = chain.from_iterable([self.id2word[tokenid]] * int(cnt) for tokenid, cnt in doc)
            else:
                tokens = chain.from_iterable([str(tokenid)] * int(cnt) for tokenid, cnt in doc)
            file_like.write(utils.to_utf8("%s 0 %s\n" % (docno, ' '.join(tokens))))

    def convert_input(self, corpus, infer=False, serialize_corpus=True):
        """Convert corpus to Mallet format and save it to a temporary text file.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Collection of texts in BoW format.
        infer : bool, optional
            ...
        serialize_corpus : bool, optional
            ...

        """
        if serialize_corpus:
            logger.info("serializing temporary corpus to %s", self.text_corpus_filename())
            with utils.open(self.text_corpus_filename(), 'wb') as fout:
                self.corpus2mallet(corpus, fout)

        # convert the text file above into MALLET's internal format
        cmd = (
            self.mallet_path + " import-file --preserve-case --keep-sequence "
            "--remove-stopwords --token-regex \"\\S+\" --input %s --output %s"
        )
        if infer:
            cmd += ' --use-pipe-from ' + self.mallet_corpus_filename()
            cmd = cmd % (self.text_corpus_filename(), self.mallet_corpus_filename() + '.infer')
        else:
            cmd = cmd % (self.text_corpus_filename(), self.mallet_corpus_filename())
        logger.info("converting temporary corpus to MALLET format with %s", cmd)
        check_output(args=cmd, shell=True)

    def train(self, corpus, **_):
        """Train Mallet LDA.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int)
            Corpus in BoW format

        MALLET infer-topics --help TRUE

        A tool for estimating, saving and printing diagnostics for topic models, such as LDA.
        --help TRUE|FALSE
            Print this command line option usage information.  Give argument of TRUE for longer documentation
            Default is false
        --prefix-code 'JAVA CODE'
            Java code you want run before any other interpreted code.  Note that the text is interpreted without modification,
            so unlike some other Java code options, you need to include any necessary 'new's when creating objects.
            Default is null
        --config FILE
            Read command option values from a file
            Default is null
        --input FILENAME
            The filename from which to read the list of training instances.  Use - for stdin.
            The instances must be FeatureSequenceor FeatureSequenceWithBigrams, not FeatureVector
            Default is null
        --input-model FILENAME
            The filename from which to read the binary topic model. The --input option is ignored.
            By default this is null, indicating that no file will be read.
            Default is null
        --input-state FILENAME
            The filename from which to read the gzipped Gibbs sampling state created by --output-state.
            The original input file must be included, using --input. By default this is null, indicating that no file will be read.
            Default is null
        --output-model FILENAME
            The filename in which to write the binary topic model at the end of the iterations.
            By default this is null,indicating that no file will be written.
            Default is null
        --output-state FILENAME
            The filename in which to write the Gibbs sampling state after at the end of the iterations.
            By default this is null, indicating that no file will be written.
            Default is null
        --output-model-interval INTEGER
            The number of iterations between writing the model (and its Gibbs sampling state) to a binary file.
            You must also set the --output-model to use this option, whose argument will be the prefix of the filenames.
            Default is 0
        --output-state-interval INTEGER
            The number of iterations between writing the sampling state to a text file.
            You must also set the --output-state to use this option, whose argument will be the prefix of the filenames.
            Default is 0
        --inferencer-filename FILENAME
            A topic inferencer applies a previously trained topic model to new documents.
            By default this is null, indicating that no file will be written.
            Default is null
            --evaluator-filename FILENAME
            A held-out likelihood evaluator for new documents.
            By default this is null, indicating that no file will be written.
            Default is null
        --output-topic-keys FILENAME
            The filename in which to write the top words for each topic and any Dirichlet parameters.
            By default this is null, indicating that no file will be written.
            Default is null
        --num-top-words INTEGER
            The number of most probable words to print for each topic after model estimation.
            Default is 20
        --show-topics-interval INTEGER
            The number of iterations between printing a brief summary of the topics so far.
            Default is 50
        --topic-word-weights-file FILENAME
            The filename in which to write unnormalized weights for every topic and word type.
            By default this is null, indicating that no file will be written.
            Default is null
        --word-topic-counts-file FILENAME
            The filename in which to write a sparse representation of topic-word assignments.
            By default this is null, indicating that no file will be written.
            Default is null
        --diagnostics-file FILENAME
            The filename in which to write measures of topic quality, in XML format.
            By default this is null, indicating that no file will be written.
            Default is null
        --xml-topic-report FILENAME
            The filename in which to write the top words for each topic and any Dirichlet parameters in XML format.
            By default this is null, indicating that no file will be written.
            Default is null
        --xml-topic-phrase-report FILENAME
            The filename in which to write the top words and phrases for each topic and any Dirichlet parameters in XML format.
            By default this is null, indicating that no file will be written.
            Default is null
        --output-topic-docs FILENAME
            The filename in which to write the most prominent documents for each topic, at the end of the iterations.
            By default this is null, indicating that no file will be written.
            Default is null
        --num-top-docs INTEGER
            When writing topic documents with --output-topic-docs, report this number of top documents.
            Default is 100
        --output-doc-topics FILENAME
            The filename in which to write the topic proportions per document, at the end of the iterations.
            By default this is null, indicating that no file will be written.
            Default is null
        --doc-topics-threshold DECIMAL
            When writing topic proportions per document with --output-doc-topics, do not print topics with proportions less than this threshold value.
            Default is 0.0
        --doc-topics-max INTEGER
            When writing topic proportions per document with --output-doc-topics, do not print more than INTEGER number of topics.
            A negative value indicates that all topics should be printed.
            Default is -1
        --num-topics INTEGER
            The number of topics to fit.
            Default is 10
        --num-threads INTEGER
            The number of threads for parallel training.
            Default is 1
        --num-iterations INTEGER
            The number of iterations of Gibbs sampling.
            Default is 1000
        --num-icm-iterations INTEGER
            The number of iterations of iterated conditional modes (topic maximization).
            Default is 0
        --no-inference true|false
            Do not perform inference, just load a saved model and create a report. Equivalent to --num-iterations 0.
            Default is false
        --random-seed INTEGER
            The random seed for the Gibbs sampler.  Default is 0, which will use the clock.
            Default is 0
        --optimize-interval INTEGER
            The number of iterations between reestimating dirichlet hyperparameters.
            Default is 0
        --optimize-burn-in INTEGER
            The number of iterations to run before first estimating dirichlet hyperparameters.
            Default is 200
        --use-symmetric-alpha true|false
            Only optimize the concentration parameter of the prior over document-topic distributions.
            This may reduce the number of very small, poorly estimated topics, but may disperse common words over several topics.
            Default is false
        --alpha DECIMAL
            SumAlpha parameter: sum over topics of smoothing over doc-topic distributions. alpha_k = [this value] / [num topics]
            Default is 5.0
        --beta DECIMAL
            Beta parameter: smoothing parameter for each topic-word. beta_w = [this value]
            Default is 0.01

        """
        self.convert_input(corpus, infer=False)
        cmd = (
            self.mallet_path + ' train-topics --input %s --num-topics %s  --alpha %s --optimize-interval %s '
            '--num-threads %s --output-state %s --output-doc-topics %s --output-topic-keys %s --num-top-words %s '
            '--num-iterations %s --inferencer-filename %s --doc-topics-threshold %s  --random-seed %s'
        )

        cmd = cmd % (
            self.mallet_corpus_filename(),
            self.num_topics,
            self.alpha,
            self.optimize_interval,
            self.workers,
            self.mallet_state_filename(),
            self.document_topics_filename(),
            self.topic_keys_filename(),
            self.num_top_words,
            self.iterations,
            self.inferencer_filename(),
            self.topic_threshold,
            str(self.random_seed),
        )
        # NOTE "--keep-sequence-bigrams" / "--use-ngrams true" poorer results + runs out of memory
        logger.info("training MALLET LDA with %s", cmd)
        check_output(args=cmd, shell=True)
        self.word_topics = self.load_word_topics()
        # NOTE - we are still keeping the wordtopics variable to not break backward compatibility.
        # word_topics has replaced wordtopics throughout the code;
        # wordtopics just stores the values of word_topics when train is called.
        self.wordtopics = self.word_topics

    def __getitem__(self, bow, iterations=100):
        """Get vector for document(s).

        Parameters
        ----------
        bow : {list of (int, int), iterable of list of (int, int)}
            Document (or corpus) in BoW format.
        iterations : int, optional
            Number of iterations that will be used for inferring.

        Returns
        -------
        list of (int, float)
            LDA vector for document as sequence of (topic_id, topic_probability) **OR**
        list of list of (int, float)
            LDA vectors for corpus in same format.

        """
        is_corpus, corpus = utils.is_corpus(bow)
        if not is_corpus:
            # query is a single document => make a corpus out of it
            bow = [bow]

        self.convert_input(bow, infer=True)
        cmd = (
            self.mallet_path + ' infer-topics --input %s --inferencer %s '
            '--output-doc-topics %s --num-iterations %s --doc-topics-threshold %s --random-seed %s'
        )
        cmd = cmd % (
            self.mallet_corpus_filename() + '.infer',
            self.inferencer_filename(),
            self.document_topics_filename() + '.infer',
            iterations,
            self.topic_threshold,
            str(self.random_seed),
        )
        logger.info("inferring topics with MALLET LDA '%s'", cmd)
        check_output(args=cmd, shell=True)
        result = list(self.read_doctopics(self.document_topics_filename() + '.infer'))
        return result if is_corpus else result[0]

    def load_word_topics(self):
        """Load words X topics matrix from :meth:`gensim.models.wrappers.ldamallet.LdaMallet.mallet_state_filename` file.

        Returns
        -------
        numpy.ndarray
            Matrix words X topics.

        """
        logger.info("loading assigned topics from %s", self.mallet_state_filename())
        word_topics = numpy.zeros((self.num_topics, self.num_terms), dtype=numpy.float64)
        if hasattr(self.id2word, 'token2id'):
            word2id = self.id2word.token2id
        else:
            word2id = revdict(self.id2word)

        with utils.open(self.mallet_state_filename(), 'rb') as fp:
            _ = next(fp)  # header
            self.alpha = numpy.fromiter(next(fp).split()[2:], dtype=float)
            assert len(self.alpha) == self.num_topics, "mismatch between MALLET vs. requested topics"
            _ = next(fp)  # noqa:F841 beta
            for lineno, line in enumerate(fp):
                line = utils.to_unicode(line)
                doc, source, pos, typeindex, token, topic = line.split(" ")
                if token not in word2id:
                    continue
                tokenid = word2id[token]
                word_topics[int(topic), tokenid] += 1.0
        return word_topics

    def load_document_topics(self):
        """Load document topics from :meth:`gensim.models.wrappers.ldamallet.LdaMallet.document_topics_filename` file.
        Shortcut for :meth:`gensim.models.wrappers.ldamallet.LdaMallet.read_doctopics`.

        Returns
        -------
        iterator of list of (int, float)
            Sequence of LDA vectors for documents.

        """
        return self.read_doctopics(self.document_topics_filename())

    def get_topics(self):
        """Get topics X words matrix.

        Returns
        -------
        numpy.ndarray
            Topics X words matrix, shape `num_topics` x `vocabulary_size`.

        """
        topics = self.word_topics
        return topics / topics.sum(axis=1)[:, None]

    def show_topics(
        self, num_topics: int = 10, num_words: int = 10, log: bool = False
    ) -> list[tuple[int, list[str, float]]]:
        """Get the `num_words` most probable words for `num_topics` number of topics.

        Parameters
        ----------
        num_topics : int, optional
            Number of topics to return, set `-1` to get all topics.
        num_words : int, optional
            Number of words.
        log : bool, optional
            If True - write topic with logging too, used for debug proposes.

        Returns
        -------
        list of (float, str)
            Topics as list of (weight, word) pairs

        """
        if num_topics < 0 or num_topics >= self.num_topics:
            num_topics = self.num_topics
            chosen_topics = range(num_topics)
        else:
            num_topics = min(num_topics, self.num_topics)
            # add a little random jitter, to randomize results around the same alpha
            sort_alpha = self.alpha + 0.0001 * numpy.random.rand(len(self.alpha))
            sorted_topics = list(matutils.argsort(sort_alpha))
            chosen_topics = sorted_topics[: num_topics // 2] + sorted_topics[-num_topics // 2 :]
        shown = []
        for i in chosen_topics:
            topic = self.show_topic(i, topn=num_words)
            shown.append((i, topic))
            if log:
                logger.info("topic #%i (%.3f): %s", i, self.alpha[i], topic)
        return shown

    def show_topic(self, topicid, topn=10, num_words=None):
        """Get `num_words` most probable words for the given `topicid`.

        Parameters
        ----------
        topicid : int
            Id of topic.
        topn : int, optional
            Top number of topics that you'll receive.
        num_words : int, optional
            DEPRECATED PARAMETER, use `topn` instead.

        Returns
        -------
        list of (str, float)
            Sequence of probable words, as a list of `(word, word_probability)` for `topicid` topic.

        """
        if num_words is not None:  # deprecated num_words is used
            warnings.warn("The parameter `num_words` is deprecated, will be removed in 4.0.0, use `topn` instead.")
            topn = num_words

        if self.word_topics is None:
            logger.warning("Run train or load_word_topics before showing topics.")
        topic = self.word_topics[topicid]
        topic = topic / topic.sum()  # normalize to probability dist
        bestn = matutils.argsort(topic, topn, reverse=True)
        beststr = [(self.id2word[idx], topic[idx]) for idx in bestn]
        return beststr

    def get_version(self, mallet_folder: str) -> str:
        """Get the version of Mallet."""
        try:
            archive = zipfile.ZipFile(mallet_folder, 'r')
            if u'cc/mallet/regression/' not in archive.namelist():
                return '2.0.7'
            else:
                return '2.0.8RC3'
        except Exception:
            xml_path = 'bin'.join(mallet_folder.split("bin")[:-1])
            try:
                doc = et.parse(xml_path + "pom.xml").getroot()
                namespace = doc.tag[: doc.tag.index('}') + 1]
                return doc.find(namespace + 'version').text.split("-")[0]
            except Exception:
                return "Can't parse pom.xml version file"

    def read_doctopics(self, fname, eps=1e-6, renorm=True):
        """Get document topic vectors from MALLET's "doc-topics" format, as sparse gensim vectors.

        Parameters
        ----------
        fname : str
            Path to input file with document topics.
        eps : float, optional
            Threshold for probabilities.
        renorm : bool, optional
            If True - explicitly re-normalize distribution.

        Raises
        ------
        RuntimeError
            If any line in invalid format.

        Yields
        ------
        list of (int, float)
            LDA vectors for document.

        """
        mallet_version = self.get_version(self.mallet_path) or ""
        if mallet_version.startswith("2.0.7"):
            raise RuntimeError("MALLET vesion 2.0.7 not supported, please upgrade")

        every_second_int_and_float: bool = None

        with utils.open(fname, 'rb') as fin:
            for lineno, line in enumerate(fin):
                if lineno == 0 and line.startswith(b"#doc "):
                    continue

                parts = line.split()[2:]  # skip "doc" and "source" columns

                if every_second_int_and_float is None:
                    every_second_int_and_float = not any('.' in str(x) for x in parts[0::2]) and all(
                        '.' in str(x) for x in parts[1::2]
                    )

                if every_second_int_and_float or (len(parts) == 2 * self.num_topics):
                    doc = [
                        (int(id_), float(weight)) for id_, weight in zip(*[iter(parts)] * 2) if abs(float(weight)) > eps
                    ]
                elif len(parts) == self.num_topics:
                    doc = [(id_, float(weight)) for id_, weight in enumerate(parts) if abs(float(weight)) > eps]
                else:
                    raise RuntimeError("invalid doc topics format at line %i in %s" % (lineno + 1, fname))

                if renorm:
                    # explicitly normalize weights to sum up to 1.0, just to be sure...
                    total_weight = float(sum(weight for _, weight in doc))
                    if total_weight:
                        doc = [(id_, float(weight) / total_weight) for id_, weight in doc]
                yield doc

    @classmethod
    def load(cls, *args, **kwargs):
        """Load a previously saved LdaMallet class. Handles backwards compatibility from
        older LdaMallet versions which did not use random_seed parameter.
        """
        model = super(LdaMallet, cls).load(*args, **kwargs)
        if not hasattr(model, 'random_seed'):
            model.random_seed = 0

        return model


def malletmodel2ldamodel(mallet_model, gamma_threshold=0.001, iterations=50):
    """Convert :class:`~gensim.models.wrappers.ldamallet.LdaMallet` to :class:`~gensim.models.ldamodel.LdaModel`.

    This works by copying the training model weights (alpha, beta...) from a trained mallet model into the gensim model.

    Parameters
    ----------
    mallet_model : :class:`~gensim.models.wrappers.ldamallet.LdaMallet`
        Trained Mallet model
    gamma_threshold : float, optional
        To be used for inference in the new LdaModel.
    iterations : int, optional
        Number of iterations to be used for inference in the new LdaModel.

    Returns
    -------
    :class:`~gensim.models.ldamodel.LdaModel`
        Gensim native LDA.

    """
    model_gensim = LdaModel(
        id2word=mallet_model.id2word,
        num_topics=mallet_model.num_topics,
        alpha=mallet_model.alpha,
        eta=0,
        iterations=iterations,
        gamma_threshold=gamma_threshold,
        dtype=numpy.float64,  # don't loose precision when converting from MALLET
    )
    model_gensim.state.sstats[...] = mallet_model.wordtopics
    model_gensim.sync_state()
    return model_gensim
